//=============================================================================
//		opnaw : OPNA unit with wait
//			ver 0.03
//=============================================================================

#include	<windows.h>
#include	<stdio.h>
#include	<string.h>
#include	<math.h>
#include	"opnaw.h"
#include	"ppz8l.h"
#include	"util.h"
#include	"misc.h"



// ---------------------------------------------------------------------------
//	̃jbgŐ`Ԃsꍇɐ錾
//	fmgen 007 Ő`Ԃp~ꂽߒǉ
//-----------------------------------------------------------------------------
#define INTERPOLATION_IN_THIS_UNIT


namespace FM {

//-----------------------------------------------------------------------------
//	RXgN^EfXgN^
//-----------------------------------------------------------------------------
OPNAW::OPNAW()
{
	fmwait = 0;
	ssgwait = 0;
	rhythmwait = 0;
	adpcmwait = 0;
	fmwaitcount = 0;
	ssgwaitcount = 0;
	rhythmwaitcount = 0;
	adpcmwaitcount = 0;
	interpolation2 = false;
	
	read_pos = 0;
	write_pos = 0;
	rate2 = 0;
	count2 = 0;
	delta = 0;
	memset(pre_buffer, 0, sizeof(pre_buffer));
	memset(ip_buffer, 0, sizeof(ip_buffer));
	m_S98 = NULL;
}

OPNAW::~OPNAW()
{
}


//-----------------------------------------------------------------------------
//	
//-----------------------------------------------------------------------------
bool OPNAW::Init(uint c, uint r, bool ipflag, const char* path, S98File* s98)
{
	memset(pre_buffer, 0, sizeof(pre_buffer));
	memset(ip_buffer, 0, sizeof(ip_buffer));
	read_pos = 0;
	write_pos = 0;
	count2 = 0;
	delta = 0;
	rate2 = r;
	m_S98 = s98;

#ifdef INTERPOLATION_IN_THIS_UNIT
	if(ipflag) {
		return OPNA::Init(c, SOUND_55K, false, path);
	} else {
		return OPNA::Init(c, r, false, path);
	}
#else
	return OPNA::Init(c, r, ipflag, path);
#endif
}

//-----------------------------------------------------------------------------
//	TvO[gύX
//-----------------------------------------------------------------------------
bool OPNAW::SetRate(uint c, uint r, bool ipflag)
{
	SetFMWait(fmwait);
	SetSSGWait(ssgwait);
	SetRhythmWait(rhythmwait);
	SetADPCMWait(adpcmwait);
	
	interpolation2 = ipflag;
	rate2 = r;
	
#ifdef INTERPOLATION_IN_THIS_UNIT
	if(ipflag) {
		return OPNA::SetRate(c, SOUND_55K, false);
	} else {
		return OPNA::SetRate(c, r, false);
	}
#else
	return OPNA::SetRate(c, r, ipflag);
#endif
	fmwaitcount = fmwait * rate / 1000000;
	ssgwaitcount = ssgwait * rate / 1000000;
	rhythmwaitcount = rhythmwait * rate / 1000000;
	adpcmwaitcount = adpcmwait * rate / 1000000;
}

//-----------------------------------------------------------------------------
//	Wait ݒ
//-----------------------------------------------------------------------------
void OPNAW::SetFMWait(int nsec)
{
	fmwait = nsec;
	fmwaitcount = nsec * rate / 1000000;
}

void OPNAW::SetSSGWait(int nsec)
{
	ssgwait = nsec;
	ssgwaitcount = nsec * rate / 1000000;
}

void OPNAW::SetRhythmWait(int nsec)
{
	rhythmwait = nsec;
	rhythmwaitcount = nsec * rate / 1000000;
}

void OPNAW::SetADPCMWait(int nsec)
{
	adpcmwait = nsec;
	adpcmwaitcount = nsec * rate / 1000000;
}

//-----------------------------------------------------------------------------
//	Wait 擾
//-----------------------------------------------------------------------------
int OPNAW::GetFMWait(void)
{
	return fmwait;
}

int OPNAW::GetSSGWait(void)
{
	return ssgwait;
}

int OPNAW::GetRhythmWait(void)
{
	return rhythmwait;
}

int OPNAW::GetADPCMWait(void)
{
	return adpcmwait;
}

// ---------------------------------------------------------------------------
//	WX^ACɃf[^ݒ
//-----------------------------------------------------------------------------
void OPNAW::SetReg(uint addr, uint data)
{
	m_S98->SetReg(addr, data);

	if(addr < 0x10) {					// SSG
		if(ssgwaitcount) {
			CalcWaitPCM(ssgwaitcount);
		}
	} else if((addr % 100) <= 0x10) {	// ADPCM
		if(adpcmwaitcount) {
			CalcWaitPCM(adpcmwaitcount);
		}
	} else if(addr < 0x20) {			// RHYTHM
		if(rhythmwaitcount) {
			CalcWaitPCM(rhythmwaitcount);
		}
	} else {							// FM
		if(fmwaitcount) {
			CalcWaitPCM(fmwaitcount);
		}
	}
	
	OPNA::SetReg(addr, data);
}

// ---------------------------------------------------------------------------
//	SetReg() wait  PCM vZ
//-----------------------------------------------------------------------------
void OPNAW::CalcWaitPCM(int waitcount)
{
	int		outsamples;

	count2 += waitcount % 1000;
	waitcount /= 1000;
	if(count2 > 1000) {
		waitcount++;
		count2 -= 1000;
	}

	do {
		if(write_pos + waitcount > WAIT_PCM_BUFFER_SIZE) {
			outsamples = WAIT_PCM_BUFFER_SIZE - write_pos;
		} else {
			outsamples = waitcount;
		}
		
		memset(&pre_buffer[write_pos * 2], 0, outsamples * 2 * sizeof(Sample));
		OPNA::Mix(&pre_buffer[write_pos * 2], outsamples);
		write_pos += outsamples;
		if(write_pos == WAIT_PCM_BUFFER_SIZE) {
			write_pos = 0;
		}
		waitcount -= outsamples;
	} while (waitcount > 0);
}


// ---------------------------------------------------------------------------
//	iꎟԂȂVer.)
//-----------------------------------------------------------------------------
void OPNAW::_Mix(Sample* buffer, int nsamples)
{
	int		bufsamples;
	int		outsamples;
	int		i;

	if(read_pos != write_pos) {			// buffer o
		if(read_pos < write_pos) {
			bufsamples = write_pos - read_pos;
		} else {
			bufsamples = write_pos - read_pos + WAIT_PCM_BUFFER_SIZE;
		}
		if(bufsamples > nsamples) {
			bufsamples = nsamples;
		}
		
		do {
			if(read_pos + bufsamples > WAIT_PCM_BUFFER_SIZE) {
				outsamples = WAIT_PCM_BUFFER_SIZE - read_pos;
			} else {
				outsamples = bufsamples;
			}
			
			for(i = 0; i < outsamples * 2; i++) {
				*buffer += pre_buffer[read_pos * 2 + i];
				buffer++;
			}

//			memcpy(buffer, &pre_buffer[read_pos * 2], outsamples * 2 * sizeof(Sample));
			read_pos += outsamples;
			if(read_pos == WAIT_PCM_BUFFER_SIZE) {
				read_pos = 0;
			}
			nsamples -= outsamples;
			bufsamples -= outsamples;
		} while (bufsamples > 0);
	}
	
	OPNA::Mix(buffer, nsamples);
}

// ---------------------------------------------------------------------------
//	
//-----------------------------------------------------------------------------
void OPNAW::Mix(Sample* buffer, int nsamples)
{

#ifdef INTERPOLATION_IN_THIS_UNIT
	int	nmixdata2;

	if(interpolation2) {
		while(nsamples > 0) {
			int nmixdata = min((delta + (nsamples - 1) * SOUND_55K / rate2 * 16384) / 16384 + 1, IP_PCM_BUFFER_SIZE - 1);
			memset(&ip_buffer[2], 0, sizeof(Sample) * 2 * nmixdata);
			_Mix(&ip_buffer[2], nmixdata);

			nmixdata2 = 0;
			while(nmixdata > nmixdata2) {
				*buffer++ += (ip_buffer[nmixdata2*2    ] * (16384 - delta) + ip_buffer[nmixdata2*2+2] * delta) / 16384;
				*buffer++ += (ip_buffer[nmixdata2*2 + 1] * (16384 - delta) + ip_buffer[nmixdata2*2+3] * delta) / 16384;
				delta += SOUND_55K * 16384 / rate2;
				nmixdata2 += delta / 16384;
				delta %= 16384;
				nsamples--;
			}
			ip_buffer[0] = ip_buffer[nmixdata * 2    ];
			ip_buffer[1] = ip_buffer[nmixdata * 2 + 1]; 
		}
	} else {
		_Mix(buffer, nsamples);
	}
#else
	_Mix(buffer, nsamples);
#endif
}

	
// ---------------------------------------------------------------------------
//	obt@NA
//-----------------------------------------------------------------------------
void OPNAW::ClearBuffer(void)
{
	read_pos = write_pos = 0;
	memset(pre_buffer, 0, sizeof(pre_buffer));
	memset(ip_buffer, 0, sizeof(ip_buffer));
}

}