#include <stdlib.h>
#include <string.h>
#include "s98file.h"
#pragma warning(disable : 4996)

S98File::S98File() :
	m_fp(NULL),
	m_start(false),
	m_wait(0)
{
}

S98File::~S98File()
{
	if (m_fp != NULL) {
		Close();
		m_fp = NULL;
	}
}

static void SetDwordToLE(unsigned char *_ptr, unsigned int _dat)
{
	_ptr[0x0] = (_dat    ) & 0xff;
	_ptr[0x1] = (_dat>> 8) & 0xff;
	_ptr[0x2] = (_dat>>16) & 0xff;
	_ptr[0x3] = (_dat>>24) & 0xff;
}

bool S98File::Open(char *_path)
{
	static const Uint8 S98HEAD[] = {
	//	00   01   02   03   04   05   06   07   08   09   0A   0B   0C   0D   0E   0F
		0x53,0x39,0x38,0x33,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x20,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
	};

	if (m_fp != NULL) {
		Close();
	}

	char outname[_MAX_PATH];
	{
		char drive[_MAX_DRIVE];
		char dir[_MAX_DIR];
		char fname[_MAX_FNAME];
		char ext[_MAX_EXT];
		_splitpath( _path, drive, dir, fname, ext );
		strcpy( outname, fname );
		strcat( outname, ".s98" );
	}

	// fopen
	m_fp = fopen(outname, "w+b");
	if (m_fp == NULL) {
		return false;
	}

	Uint8 header[0x20];
	memcpy(header, S98HEAD, 0x20);

	SetDwordToLE(&header[0x04], 1);
	SetDwordToLE(&header[0x08], 1000);

	fwrite(header, sizeof(Uint8), sizeof(header), m_fp);

	return true;
}

void S98File::Close()
{
	Stop();

	if (m_fp != NULL) {
		Sync(0, true);
		Uint8 eod[] = {0xfd};
		fwrite(eod, sizeof(Uint8), 1, m_fp);
		fclose(m_fp);
		m_fp = NULL;
	}
}

void S98File::Start()
{
	m_start = true;
	m_wait = 0.0;
}

void S98File::Stop()
{
	m_start = false;
}

void S98File::SetReg(Uint32 _reg, Uint32 _data)
{
	if (m_start == false) {
		return;
	}

	if (m_fp) {
		Uint8 data[3];
		data[0] = (_reg >= 0x100) ? 0x01 : 0x00;
		data[1] = _reg & 0xff;
		data[2] = _data & 0xff;
		fwrite(data, sizeof(Uint8), sizeof(data), m_fp);
	}
}

void S98File::Sync(double _num, bool _force)
{
	m_wait += _num;
	int iwait = static_cast<int>(m_wait);
	m_wait -= iwait;

	Uint8 buf[2] = { 0xff, 0xff };
	if (m_fp) {
		switch(iwait) {
		  case 0:
			// do nothing
			break;
		  case 1:
			fwrite(buf, sizeof(Uint8), 1, m_fp);
			break;
		  case 2:
			fwrite(buf, sizeof(Uint8), 2, m_fp);
			break;
		  default:
			buf[0] = 0xfe;
			fwrite(buf, sizeof(Uint8), 1, m_fp);
			iwait -= 2;
			while  (iwait > 0x7f)
			{
				buf[0] = 0x80 | (iwait & 0x7f);
				fwrite(buf, sizeof(Uint8), 1, m_fp);
				iwait >>= 7;
			}
			buf[0] = iwait & 0x7f;
			fwrite(buf, sizeof(Uint8), 1, m_fp);
			break;
		}
	}
}
